#!/usr/bin/python
#############################################################################
## eeMark source code generator                                            ##
## http://eeclust.de                                                       ##
##                                                                         ##
## Copyright (c) 2011-2012, ZIH, TU Dresden, Federal Republic of Germany   ##
##                                                                         ##
## See the file COPYING in the package base directory for details          ##
#############################################################################
import os, stat, sys, re, ConfigParser, getopt, subprocess
from Cheetah.Template import Template

###
# Declarations and initialisation:
##

configFile = 'config.cfg'
modesFileName = 'eeClust_intel_modes.cfg'
benchspecFile = 'simple.benchspec'
debug = True

cfg = ConfigParser.ConfigParser()
#benchspec = ConfigParser.ConfigParser()

funcList = []
#rankmap = {}
prefix = ''
settings = {}
pragmas = {}
action='source'
valid_actions=('source', 'build', 'run', 'buildrun')
class data_types:
	base = ['int32', 'int64', 'single', 'double']
	c = {'int32':'uint32_t', 'int64':'uint64_t', 'single':'float', 'double':'double'}
	wrkld = {'int32':'iops_32', 'int64':'iops_64', 'single':'flops_sp', 'double':'flops_dp'}
	mpi = {'int32' : 'MPI_INT', 'int64' : 'MPI_LONG_LONG', 'single' : 'MPI_FLOAT', 'double' : 'MPI_DOUBLE'}
	width = {'int32' : 4, 'int64' : 8, 'single' : 4, 'double' : 8}
sum_fixed = 0
sum_dynamic = 0
num_groups = 0

###
# Functions:
##

# Exits with given error message
def exitWithError(msg):
	print "ERROR:", msg
	sys.exit(1)

# reads all section names from the config file
# single means: do not return a split list, but a single value
def getCfgOpt(config, section, name, single=False):
	if config.has_option(section,name):
		if single:
			return config.get(section,name).strip()
		else:
			return map(lambda x: x.strip(), config.get(section,name).split(','))
	else:
		if single: return ''
		else: return []


###
# Command line arguments
##

try:
	opts,args = getopt.getopt(sys.argv[1:], "c:a:h", ["config=", "action=", "help"])
except getopt.GetoptError, err:
	exitWithError(str(err))
for o, a in opts:
	if o in ("-c", "--config"):
		configFile=a
	elif o in ("-b", "--benchspec"):
		benchspecFile=a
	elif o in ("-a", "--action"):
		action=a.lower()
		if action not in valid_actions:
			exitWithError("invalid action. Use --action {"+"|".join(valid_actions)+"}")
	elif o in ("-h", "--help"):
		print "Possible arguments:"
		print " -a|--action {"+"|".join(valid_actions)+"}"
                print " -b|--benchspec <benchspecfile>"
		print " -c|--config <configfile>"
		print " -h|--help"
		sys.exit(0)
	if action.find('run') != -1:
		if benchspecFile == '':
			exitWithError("action run requires benchspec. Use -b to specify benchspec file")

# load config file
if not os.path.exists(configFile):
	exitWithError("config file \""+configFile+"\" doesn't exist. Use --config <configfile>")
else:
	cfg.readfp(open(configFile))
        print "using config: \""+configFile+"\""

'''
# load benchspec file
if not os.path.exists(benchspecFile):
	exitWithError("benchmark specification file \""+benchspecFile+"\" doesn't exist. Use --benchspec <benchspec>")
else:
	benchspec.readfp(open(benchspecFile))
        print "using benchmark specification: \""+benchspecFile+"\""
'''

###
# Main program begins here:
##

# default settings
settings['add_cpu_instrumentation'] = 'no'
settings['add_instrumentation'] = 'no'
settings['ee_mode_cpu_default'] = 'MODE_MAX'
settings['ee_mode_disk_default'] = 'MODE_MAX'
settings['ee_mode_nic_default'] = 'MODE_MAX'
settings['create_pbs'] = 'no'

# process the config file
for each in cfg.sections():
	if each == 'general':
		for opt in cfg.options(each):
			if opt == 'prefix':
				prefix = getCfgOpt(cfg, each, opt, True)
				if prefix: prefix = prefix + '_'
			else:
				settings[opt] = getCfgOpt(cfg, each, opt, True)
	elif each == 'pragma':
		for opt in cfg.options(each):
			pragmas[opt.upper()] = getCfgOpt(cfg, each, opt, True).strip('"')
	else: # functions
		mo=re.match("func_(\w+)", each)
		if mo!=None:
			name=mo.group(1).lower()
			mopb = getCfgOpt(cfg, each, "max_ops_per_byte", True) or '1'
			initTmpl = getCfgOpt(cfg, each, "init", True) or name
			funcTmpl = getCfgOpt(cfg, each, "tmpl", True) or name
			datatypeList = getCfgOpt(cfg, each, "datatypes", False) or data_types.base
			tmp = getCfgOpt(cfg, each, "ops_per_byte_list", False)
			if not tmp:
				opbList = range(1,int(mopb)+1)
			else:
				opbList = []
				# parse list for ranges etc.:
				for value in tmp:
					mo = re.match('(\d+)-(\d+):(\d+)', value) # range with step
					if mo!=None:
						for i in range(int(mo.group(1)), int(mo.group(2))+1, int(mo.group(3))):
							if i>int(mopb): mopb=str(i);
							if i not in opbList: opbList.append(i)			
						continue
					mo = re.match('(\d+)-(\d+)', value) # simple range
					if mo!=None:
						for i in range(int(mo.group(1)), int(mo.group(2))+1):
							if i>int(mopb): mopb=str(i);
							if i not in opbList: opbList.append(i)
						continue
					if int(value)>int(mopb): mopb=value;
					if int(value) not in opbList: opbList.append(int(value)) # single values

			tmp = [name, mopb, getCfgOpt(cfg, each, "type", True), initTmpl, funcTmpl, opbList, datatypeList]
			funcList.append(tmp)

#add libraries for powertracer energylib
if settings['energylib'] == 'powertracer':
	settings['ldflags'] += ' `pkg-config --libs glib-2.0 gthread-2.0` -lpq'

funcList.sort();

modes = {}
# read modes from modeFileName and create modes[func][datatype][opb] with eeMode entries...
if settings['add_cpu_instrumentation'] == 'yes':	

	modeFile = open(modesFileName, "r")
	lineNo = 1;
	for line in modeFile:
		
		#skip empty lines and comments
		if line == "" or line[0] == "#":
			lineNo += 1
			continue
	
		var = line.split(" ")
		func_datatype_opb = var[0].split("_")
		mode = var[1].strip()
		func = func_datatype_opb[0]
		datatype = func_datatype_opb[1]
		opb = int(func_datatype_opb[2])

		if modes.has_key(func):

			modes_func = modes[func]
			
			if modes_func.has_key(datatype):
				modes_func_datatype = modes_func[datatype]
				
				if modes_func_datatype.has_key(opb):
					# error: key already specified
					exitWithError("Double key specified in line %d: %s" %(lineNo, func_datatype_opb))
				
				modes_func_datatype[opb] = mode			
			else:
				modes_func[datatype] = {}
				modes_func[datatype][opb] = mode
		else:
			modes_func_datatype = {}
			modes_func_datatype[opb] = mode
			modes_func = {}
			modes_func[datatype] = modes_func_datatype
			modes[func] = modes_func

		lineNo+= 1

#DEBUG
if debug:
	print "settings: "
	print settings
	print "funclist: "
	print funcList
	print "modes: "
	print modes

## TODO - unused as group definition was changed to allow dynamic rank numbers (determined at runtime)
##      - might be reused for optional explicit rank placement (distribution = manual)
## # create rank map according to groups in config file
## for each in groupList:
##	ranklist = []
##	tmp = each['ranks'].split(',')
##	for ranksec in tmp:
##		ranksec = ranksec.strip()
##		mo = re.match('(\d+)-(\d+):(\d+)', ranksec) # range with step width
##		if mo!=None:
##			for i in range(int(mo.group(1)), int(mo.group(2))+1, int(mo.group(3))):
##				ranklist.append(i)			
##			continue
##		mo = re.match('(\d+)-(\d+)', ranksec) # simple range
##		if mo!=None:
##			for i in range(int(mo.group(1)), int(mo.group(2))+1):
##				ranklist.append(i)
##			continue
##		ranklist.append(int(ranksec))
##
##	#_DEBUG
##	if debug:
##		print each['name'], ranklist
##
##	for r in ranklist:
##		if r >= settings['ranks']:
##			exitWithError('Rank out of bounds ('+str(r)+') in group "'+each['name']+'"')
##		rankmap[r] = each['function']

if action.find('build') != -1 or action.find('source') != -1:
	if prefix!="":
		short_prefix = '_'+prefix[:-1]
	else:
		short_prefix = ''
	#create the subdirectories
	if not os.path.exists('src'+short_prefix):
		os.mkdir('src'+short_prefix)
	if not os.path.exists('data'):
		os.mkdir('data')
	if not os.path.exists('results'):
		os.mkdir('results')
		

	outfilename=prefix+'eeMark'

	# create the C files:
	print "Processing templates..."
	gen_files=['eeMark.c', 'eeMark.h', 'config_parser.c', 'energymeter_interface.h', 'aligned_memory.h', 'aligned_memory.c', 'init_functions.c']
	os.chdir('tmpl') # this is necessary for the imports within the templates to work
	for tf in gen_files:
		print tf+".tmpl ...",
		t = str(Template(file=tf+".tmpl", searchList = [{'prefix' : prefix, 'settings' : settings, 'funclist' : funcList, 'pragmas': pragmas, 'datatypes' : data_types}]))
		srcOutFile = "../src"+short_prefix+'/'+tf

		oldFile = ""
		if os.path.exists(srcOutFile):
			tmp = open(srcOutFile, "r")
			oldFile = tmp.read()
			tmp.close()

		if oldFile == t:
			print "skipping. not modified.",
		else:
			output = open(srcOutFile, 'w')
			output.write(t)
			output.close()
		print ""

	for each in funcList:
		print each[0]+" tmpl ...",
		t = str(Template(file="functions.c.tmpl", searchList = [{'prefix' : prefix, 'func' : each, 'settings' : settings, 'pragmas': pragmas, 'datatypes' : data_types, 'name' : each[0], 'type' : each[2], 'tmpl' : each[4], 'opbList' : each[5], 'datatypeList' : each[6], 'modes' : modes }]))
		srcOutFile = "../src"+short_prefix+"/functions_"+each[0]+".c"

		oldFile = ""
		if os.path.exists(srcOutFile):
			tmp = open(srcOutFile, "r")
			oldFile = tmp.read()
			tmp.close()

		if oldFile == t:
			print "skipping. not modified.",
		else:
			output = open(srcOutFile, 'w')
			output.write(t)
			output.close()
		print ""


	os.chdir('..')
	defines=settings['defines']
	for i,p in enumerate(pragmas):
		defines=defines+' -D'+p
	if settings['ccflags'].find('-std=c99') != -1:
		defines=defines+' -DC99'
	#remove energylib as the measurement methode could have been altered (Makefile does not notice that)
	if os.path.exists('lib/libenergyclient.a'): 
		os.remove('lib/libenergyclient.a')

	#create compile script:
	print "Creating compile.sh ..."
	output = open(prefix+"compile.sh", "w")
        output.write('#!/bin/sh\n')
	output.write('###########################################################################\n')
	output.write('# eeMark                                                                  #\n')
	output.write('# http://eeclust.de                                                       #\n')
	output.write('#                                                                         #\n')
	output.write('# Copyright (c) 2011-2012, ZIH, TU Dresden, Federal Republic of Germany   #\n')
	output.write('#                                                                         #\n')
	output.write('# See the file COPYING in the package base directory for details          #\n')
	output.write('###########################################################################\n')
        output.write('CC="'+settings['cc']+'"\n')
        output.write('CFLAGS="'+settings['ccflags']+'"\n')
        output.write('LDFLAGS="'+settings['ldflags']+'"\n')
        output.write('SRC="src'+short_prefix+'/config_parser.c src'+short_prefix+'/functions_*.c src'+short_prefix+'/init_functions.c src'+short_prefix+'/aligned_memory.c src'+short_prefix+'/'+outfilename+'.c"\n')
        output.write('TARGET='+outfilename+'\n')
	output.write('DEFINES="'+defines+' -DVERSION_DATE=`date +%m%d%Y | sed \'s/0*//\'`"\n')
	output.write('ENERGYLIB='+settings['energylib']+'\n')
	output.write('cd lib && ./${ENERGYLIB}_compile.sh && cd ..\n')
	
	output.write('echo "Compiling benchmark: $TARGET"\n')
	output.write('echo "$CC $CFLAGS $SRC $DEFINES $LDFLAGS -lenergyclient -L. -L./lib -o $TARGET"\n')
        output.write('$CC $CFLAGS $SRC $DEFINES $LDFLAGS -lenergyclient -L. -L./lib -o $TARGET\n')
	output.close()
	os.chmod(prefix+"compile.sh", 0755) #make the file executable

	#create make file
	print "Creating Makefile ..."
	output = open("Makefile", "w")
	output.write('###########################################################################\n')
	output.write('# eeMark                                                                  #\n')
	output.write('# http://eeclust.de                                                       #\n')
	output.write('#                                                                         #\n')
	output.write('# Copyright (c) 2011-2012, ZIH, TU Dresden, Federal Republic of Germany   #\n')
	output.write('#                                                                         #\n')
	output.write('# See the file COPYING in the package base directory for details          #\n')
	output.write('###########################################################################\n')
	output.write('CC = '+settings['cc']+'\n')
	if settings['add_instrumentation'] == 'yes' or settings['add_cpu_instrumentation'] == 'yes':	
		output.write('CFLAGS = '+settings['ccflags']+' -I/sw/eeDaemon/include `pkg-config --cflags glib-2.0 gthread-2.0`\n')
		output.write('LDFLAGS = '+settings['ldflags']+' -lenergyclient -L$(PWD) -L$(PWD)/lib -L/sw/eeDaemon/lib -leed_client\n')
	else:
		output.write('CFLAGS = '+settings['ccflags']+'\n')
		output.write('LDFLAGS = '+settings['ldflags']+' -lenergyclient -L$(PWD) -L$(PWD)/lib\n')		
	output.write('DEFINES = '+defines+' -DVERSION_DATE=`date +%m%d%Y | sed \'s/0*//\'`\n')
	output.write('TARGET = '+outfilename+'\n')
	output.write('SRC_DIR = src'+short_prefix+'\n')
	output.write('SRC = $(foreach dir,$(SRC_DIR),$(wildcard $(dir)/*.c))\n') 
	output.write('HEADER = $(foreach dir,$(SRC_DIR),$(wildcard $(dir)/*.h))\n')
	output.write('OBJ = $(SRC:.c=.o)\n')
	output.write('ENERGYLIB = '+settings['energylib']+'\n\n')
        output.write('$(TARGET): lib/libenergyclient.a $(OBJ) $(HEADER)\n')
        output.write('\t$(CC) $(CFLAGS) -o $@ $(OBJ) $(LDFLAGS)\n\n')
        output.write('%.o: %.c $(HEADER)\n')
        output.write('\t$(CC) $(CFLAGS) $(DEFINES) -c $< -o $@\n\n')
        output.write('lib/libenergyclient.a:\n')
        output.write('\tcd lib && ./$(ENERGYLIB)_compile.sh\n\n')
        output.write('clean:\n')
        output.write('\trm -f $(TARGET) $(SRC_DIR)/*.o\n')
	output.close()

	#create run script
	print "Creating run.sh ..."
	output = open(prefix+"run.sh", "w")
	output.write('###########################################################################\n')
	output.write('# eeMark                                                                  #\n')
	output.write('# http://eeclust.de                                                       #\n')
	output.write('#                                                                         #\n')
	output.write('# Copyright (c) 2011-2012, ZIH, TU Dresden, Federal Republic of Germany   #\n')
	output.write('#                                                                         #\n')
	output.write('# See the file COPYING in the package base directory for details          #\n')
	output.write('###########################################################################\n')
        output.write('#!/bin/sh\n')
        output.write('MPICMD="'+settings['mpicmd']+'"\n')
        output.write('NP_PARAM='+settings['np_param']+'\n')
        output.write('NPROC='+settings['np_default']+'\n')
        output.write('EXEC='+outfilename+'\n')
        output.write('${MPICMD} ${NP_PARAM} ${NPROC} ./${EXEC} $@\n')
	output.close()
	os.chmod(prefix+"run.sh", 0755) #make the file executable
	
	#create pbs file
	if settings['create_pbs'] == 'yes':
		print "Creating run.pbs ..."
		output = open(prefix+"run.pbs", "w")
		output.write('##### PBS OPTIONS ######\n')
		output.write('#PBS -q ' +settings['queue']+ '\n')
		output.write('#PBS -l nodes=' +settings['nodes'] + ':ppn='+settings['np_default']+'\n')
		output.write('#PBS -l walltime='+settings['walltime']+ '\n')
		output.write('##### END PBS OPTIONS ######\n\n')
		output.write('cd $PBS_O_WORKDIR\n\n')
	
		if settings['mpimodule'] == 'mpich2/mpich2-1.3.1-hdtrace':
			#if using hdtrace, create tracing folder...
			output.write('# create new folder with jobid\nmkdir $PBS_JOBID\n\n')
			output.write('# copy this jobfile to the created folder\ncp ${PBS_JOBNAME} ${PBS_JOBID}\n\n')
			output.write('# go into the created folder\ncd $PBS_JOBID\n\n')
			output.write('mkdir results\n\n')

		output.write('module load '+settings['mpimodule']+ '\n\n')
		if settings['energylib'] == 'powertracer' or settings['add_instrumentation'] == 'yes' or settings['add_cpu_instrumentation'] == 'yes':
			output.write('\nfor host in `cat $PBS_NODEFILE | sort -u`\ndo\n');
		if settings['energylib'] == 'powertracer':	
			output.write('\tPTLIB_NODES=$host:${PTLIB_NODES}\n')
		if settings['add_instrumentation'] == 'yes' or settings['add_cpu_instrumentation'] == 'yes':
			if settings['mpimodule'] == 'mpich2/mpich2-1.3.1-hdtrace':
				output.write('\tssh $host \"sudo /sbin/service eed-trace stop >& /dev/null\"\n')
				output.write('\tssh $host \"sudo /sbin/service eed stop >& /dev/null\"\n')
				output.write('\tssh $host \"sudo /sbin/service eed-trace start >& /dev/null\"\n')
			else:
				output.write('\tssh $host \"sudo /sbin/service eed stop >& /dev/null\"\n')
				output.write('\tssh $host \"sudo /sbin/service eed start >& /dev/null\"\n')
		if settings['energylib'] == 'powertracer' or settings['add_instrumentation'] == 'yes' or settings['add_cpu_instrumentation'] == 'yes':	
			output.write('done\n')
		if settings['energylib'] == 'powertracer':	
			output.write('\nexport PTLIB_NODES=`echo ${PTLIB_NODES} | sed \'s/\\(.*\\)./\\1/\'`\n\n')
		if settings['add_instrumentation'] == 'yes' or settings['add_cpu_instrumentation'] == 'yes':
			output.write('\nexport LD_LIBRARY_PATH=/sw/eeDaemon/lib:$LD_LIBRARY_PATH\n\n')
		output.write('# run a single benchspec\n')
		output.write('./run.sh benchspec/'+benchspecFile+'\n\n')
		output.write('# uncomment line to run a benchset (only run.sh or runset.sh is possible!)\n')
		output.write('#./runset.sh benchspec/reference.benchset\n\n')
		if settings['add_instrumentation'] == 'yes' or settings['add_cpu_instrumentation'] == 'yes':
			output.write('\nfor host in `cat $PBS_NODEFILE | sort -u`\ndo\n')
			if settings['mpimodule'] == 'mpich2/mpich2-1.3.1-hdtrace':		
				output.write('\tssh $host \"sudo /sbin/service eed-trace stop >& /dev/null\"\n\tscp ${host}:/tmp/${EXEC}_${host}_eed.stat .\n')	
			else:
				output.write('\tssh $host \"sudo /sbin/service eed stop >& /dev/null\"\n')		
			output.write('done\n\n')
		
		if settings['mpimodule'] == 'mpich2/mpich2-1.3.1-hdtrace':
			output.write('#create project file\n/home/minartz/projects/piosimhd/HDMPIwrapper/scripts/project-description-merger.py -o ${EXEC}.proj ${EXEC}*.info > /dev/null')
	
		output.close()

if action.find('build') != -1:
	# compile the C file:
	print "Compiling..."
	#subprocess.call(['sh '+prefix+'compile.sh'], shell=True)
	subprocess.call(['make -j `nproc`'], shell=True)

if action.find('run') != -1:
	# run:
	print "Running..."
	subprocess.call(['sh '+prefix+'run.sh '+benchspecFile], shell=True)

print "Done."

