#!/bin/sh
###########################################################################
# eeMark                                                                  #
# http://eeclust.de                                                       #
#                                                                         #
# Copyright (c) 2011-2012, ZIH, TU Dresden, Federal Republic of Germany   #
#                                                                         #
# See the file COPYING in the package base directory for details          #
###########################################################################

if [ $# -eq 0 ] ; then
	echo "No benchset definition file given."
	echo "Usage: $0 [options] <benchset>"
	exit 1
fi

#pass additional parameters to run.sh
ARGS=""
while [ $# -gt 1 ]; do
 ARGS=${ARGS}$1" "
 shift
done

infile=$1
filename=`basename $infile`
if [ ! -f $infile ] ; then
	echo "File does not exist: $1"
	exit 1
fi
i=0
while [ 1 ]
do
	i=$((i + 1))
	resultfile=`printf "results/eeMark_${filename}_%03d_results.txt" $i`
	[ -f $resultfile ] || break
done


printf "| %-24s | %12s | %12s | %12s | %12s | %12s |\n" "Benchmark" "Distribution" " Iterations " "Performance " " Efficiency " "  Combined  " > $resultfile
printf "| %-24s | %12s | %12s | %12s | %12s | %12s |\n" "" "" "" "   Score    " "   Score    " "   Score    " >> $resultfile
echo "-------------------------------------------------------------------------------------------------------" >> $resultfile

pstotal=0
estotal=0
cstotal=0
results=0
i=0
lines=$(wc -l < ${infile})

# one would usually do this with while read line;do ... done < ${infile}
# However, calling mpirun in run.sh destroys the input redirection after the first iteration
while [ ${i} -lt ${lines} ]; do
	line=$(sed -n `expr ${i} + 1`"p" ${infile})
        first=$(echo ${line} | cut -c 1)
	if [ ${first} != '#' ];then
	  benchspec=$(echo ${line} | cut -d' ' -f 1)
	  iterations=$(echo ${line} | cut -d' ' -f 2)
	  distribution=$(echo ${line} | cut -d' ' -f 3)

          #echo "Running ${benchspec}: sh run.sh -i ${iterations} -d ${distribution} ${ARGS} ${benchspec}.benchspec"
	  echo "Running ${iterations} iterations of ${benchspec}.benchspec (${distribution})"

	  R=$(sh run.sh -i ${iterations} -d ${distribution} ${ARGS} ${benchspec}.benchspec | grep "Writing result to") #get result file name from output
	  R=$(echo $R | cut -d : -f 2)
	  scores=$(tail -n 3 $R)

	  pscore=$(echo $scores | cut -d- -f 2 | cut -d: -f 2)
	  escore=$(echo $scores | cut -d- -f 3 | cut -d: -f 2)
	  cscore=$(echo $scores | cut -d- -f 4 | cut -d: -f 2)
	  results=$(expr $results + 1)

	  #pscore and escore have a trailing space after the cuts, thus strings are shifted back one position in format string
	  printf "| %-24s | %12s | %12s |  %12s|  %12s| %12s |\n" "${benchspec}" "${distribution}" "${iterations}" "$pscore" "$escore" "$cscore" >> $resultfile

	  if [ ! $pscore = "n/a" ] ; then
	   	pstotal=$(echo "scale=3;$pstotal+$pscore" | bc)
	  fi
	  if [ ! $escore = "n/a" ] ; then
	  	estotal=$(echo "scale=3;$estotal+$escore" | bc)
	  fi
	  if [ ! $cscore = "n/a" ] ; then
	  	cstotal=$(echo "scale=3;$cstotal+$cscore" | bc)
	  fi
	fi
        i=$(expr $i + 1)
done

export LC_NUMERIC="POSIX"  #otherwise printf expects a comma as decimal point in de_DE
pstotal=$(printf "%.2f" $(echo "scale=3;$pstotal/$results" | bc))
estotal=$(printf "%.2f" $(echo "scale=3;$estotal/$results" | bc))
cstotal=$(printf "%.2f" $(echo "scale=3;$cstotal/$results" | bc))
echo "-------------------------------------------------------------------------------------------------------" >> $resultfile
printf "| %-24s   %12s   %12s | %12s | %12s | %12s |\n" "Result:" "" "" "$pstotal" "$estotal" "$cstotal" >> $resultfile

echo "Writing to result file: ${resultfile}"

